from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
    cast
)
from eth_account.account import Account as EthAccount
from cfx_address.utils import validate_network_id
from cfx_account.signers.local import LocalAccount
from eth_utils.crypto import (
    keccak,
)
from collections.abc import (
    Mapping,
)
from cytoolz import (
    dissoc, # type: ignore
)
from hexbytes import (
    HexBytes,
)
from eth_account.datastructures import (
    # SignedMessage,
    SignedTransaction,
)
from cfx_account._utils.signing import (
    sign_transaction_dict,
)
from cfx_account._utils.transactions import (
    Transaction,
    vrs_from,
)
from cfx_address import (
    Base32Address,
    eth_eoa_address_to_cfx_hex
)
from cfx_utils.types import (
    TxParam,
    TxDict,
    HexAddress,
)
from cfx_utils.decorators import (
    combomethod,
)
from eth_keys.datatypes import (
    PrivateKey,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

class Account(EthAccount):
    
    # _default_network_id: Optional[int]=None
    w3: Optional["Web3"] = None 
    
    @combomethod
    def set_w3(self, w3: "Web3"):
        self.w3 = w3
    
    # def set_default_network_id(self, network_id: int):
    #     self._default_network_id = network_id

    @combomethod
    def from_key(
        self, private_key: Union[bytes, str, PrivateKey], network_id: Optional[int]=None
    ) -> LocalAccount:
        """
        returns a LocalAccount object

        :param str private_key: the raw private key
        :param Optional[int] network_id: target network of the account, defaults to None
        :return LocalAccount: object with methods for signing and encrypting

        >>> acct = Account.from_key(
        ... 0xb25c7db31feed9122727bf0939dc769a96564b2de4c4726d035b36ecf1e5b364)
        >>> acct.address
        '0x1ce9454909639d2d17a3f753ce7d93fa0b9ab12e'
        >>> acct.key
        HexBytes('0xb25c7db31feed9122727bf0939dc769a96564b2de4c4726d035b36ecf1e5b364')

        # These methods are also available: sign_message(), sign_transaction(), encrypt()
        # They correspond to the same-named methods in Account.*
        # but without the private key argument
        """
        key = self._parsePrivateKey(private_key)
        if network_id is not None:
            validate_network_id(network_id)
            return LocalAccount(key, self, network_id)
        if self.w3:
            w3_network_id = self.w3.cfx.chain_id
            return LocalAccount(key, self, w3_network_id)
        return LocalAccount(key, self)

    @combomethod
    def sign_transaction(
        self, transaction_dict: TxParam, private_key: Union[bytes, str, PrivateKey]
    ) -> SignedTransaction:
        """
        Sign a transaction using a local private key. Produces signature details
        and the hex-encoded transaction suitable for broadcast using
        :meth:`w3.cfx.sendRawTransaction() <web3.client.Cfx.sendRawTransaction>`.

        Create the transaction dict for a contract method with
        `my_contract.functions.my_function().buildTransaction()
        <http://web3py.readthedocs.io/en/latest/contracts.html#methods>`_

        :param TxParam transaction_dict: the transaction with keys:
          nonce, chainId, to, data, value, storageLimit, epochHeight, gas, and gasPrice.
        :param Union[bytes, str, PrivateKey] private_key: private_key to be used for signing
        :raises TypeError: transaction_dict is not a dict-like object
        :raises ValueError: transaction's from field does not match private_key
        :return SignedTransaction: an attribute dict contains various details about the signature - most
          importantly the fields: v, r, and s
    
        >>> transaction = {
                # Note that the address must be in Base32 format or native bytes:
                'to': 'cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da',
                'nonce': 1,
                'value': 1,
                'gas': 100,
                'gasPrice': 1,
                'storageLimit': 100,
                'epochHeight': 100,
                'chainId': 1
            }
        >>> key = '0xcc7939276283a32f60d2fad7d16cac972300308fe99ec98d0e63765d02e24863'
        >>> signed = Account.sign_transaction(transaction, key)
        {'hash': HexBytes('0x692a0ea530a264f4e80ce39f393233e90638ef929c8706802e15299fd0b042b9'),
            'r': 74715349327018893060702835194036838027583623083228589573427622179540208747230,
            'rawTransaction': HexBytes('0xf861dd0101649413d2ba4ed43542e7c54fbb6c5fccb9f269c1f94c016464018080a0a52f639cbed11262a7b88d0a37aef909aa7dc2c36c40689a3d52b8bd1d9482dea054f3bdeb654f73704db4cbc12451fb4c9830ef62b0f24de1a40e4b6fe10f57b2'),  # noqa: E501
            's': 38424933894051759888751352802050752143518665905311311986258635963723328477106,
            'v': 0}
        >>> w3.eth.sendRawTransaction(signed.rawTransaction)
        """
        if not isinstance(transaction_dict, Mapping):
            raise TypeError("transaction_dict must be dict-like, got %r" % transaction_dict)

        account: LocalAccount = self.from_key(private_key)
        transaction_dict = cast(TxDict, transaction_dict)
        # allow from field, *only* if it matches the private key
        if 'from' in transaction_dict:
            if Base32Address(transaction_dict['from']).hex_address == account.hex_address:
                sanitized_transaction = dissoc(transaction_dict, 'from')
            else:
                raise ValueError("transaction[from] does match key's hex address: "
                    f"from's hex address is{Base32Address(transaction_dict['from']).hex_address}, "
                    f"key's hex address is {account.hex_address}")
                
        else:
            sanitized_transaction = transaction_dict

        # sign transaction
        (
            v,
            r,
            s,
            rlp_encoded,
        ) = sign_transaction_dict(account._key_obj, sanitized_transaction)

        transaction_hash = keccak(rlp_encoded)

        return SignedTransaction(
            rawTransaction=HexBytes(rlp_encoded),
            hash=HexBytes(transaction_hash),
            r=r,
            s=s,
            v=v,
        )

    @combomethod
    def recover_transaction(self, serialized_transaction: bytes) -> HexAddress:
        """
        Get the address of the account that signed this transaction.

        :param serialized_transaction: the complete signed transaction
        :type serialized_transaction: hex str, bytes or int
        :returns: address of signer, hex-encoded & checksummed
        :rtype: str

        .. doctest:: python

            >>> raw_transaction = '0xf86a8086d55698372431831e848094f0109fc8df283027b6285cc889f5aa624eac1f55843b9aca008025a009ebb6ca057a0535d6186462bc0b465b561c94a295bdb0621fc19208ab149a9ca0440ffd775ce91a833ab410777204d5341a6f9fa91216a6f3ee2c051fea6a0428'  # noqa: E501
            >>> Account.recover_transaction(raw_transaction)
            '0x1c7536e3605d9c16a7a3d7b1898e529396a65c23'
        """
        txn_bytes = HexBytes(serialized_transaction)
        txn = Transaction.from_bytes(txn_bytes)
        recovered_address = self._recover_hash(txn[0].hash(), vrs=vrs_from(txn)) # type: ignore
        return eth_eoa_address_to_cfx_hex(recovered_address)

    @combomethod
    def create(self, extra_entropy: str='') -> LocalAccount:
        return super().create(extra_entropy)



from typing import (
    Tuple
)
from cfx_utils.types import (
    TxDict
)
from eth_keys.datatypes import (
    PrivateKey
)

from .transactions import (
    encode_transaction,
    serializable_unsigned_transaction_from_dict,
)

def sign_transaction_dict(eth_key: PrivateKey, transaction_dict: TxDict) -> Tuple[int, int, int, bytes]:
    # generate RLP-serializable transaction, with defaults filled
    unsigned_transaction = serializable_unsigned_transaction_from_dict(transaction_dict)

    transaction_hash = unsigned_transaction.hash()

    # sign with private key
    (v, r, s) = sign_transaction_hash(eth_key, transaction_hash)

    # serialize transaction with rlp
    encoded_transaction = encode_transaction(unsigned_transaction, vrs=(v, r, s))

    return (v, r, s, encoded_transaction)

def sign_transaction_hash(key: PrivateKey, transaction_hash: bytes) -> Tuple[int, int, int]:
    signature = key.sign_msg_hash(transaction_hash)
    return signature.vrs



from typing import (
    Any,
    Dict,
    Tuple
)
import rlp
from cytoolz import (
    curry, # type: ignore
    dissoc, # type: ignore
    identity, # type: ignore
    merge, # type: ignore
    partial,
    pipe, # type: ignore
)
from eth_utils.conversions import (
    to_bytes,
    to_int,
)
from eth_utils.types import (
    is_bytes,
    is_string,
)
from eth_utils.curried import (
    apply_formatters_to_dict,
    apply_one_of_formatters,
    hexstr_if_str,
)
from eth_account._utils.legacy_transactions import (
    TRANSACTION_DEFAULTS,
)
from eth_account._utils.validation import (
    VALID_EMPTY_ADDRESSES,
    is_int_or_prefixed_hexstr,
    is_none
)

from eth_rlp import (
    HashableRLP,
)
from rlp.sedes import (
    Binary,
    big_endian_int,
    binary,
)
from cfx_address import (
    Base32Address
)
from cfx_utils.types import (
    TxDict
)

UNSIGNED_TRANSACTION_FIELDS = (
    ('nonce', big_endian_int),
    ('gasPrice', big_endian_int),
    ('gas', big_endian_int),
    ('to', Binary.fixed_length(20, allow_empty=True)),
    ('value', big_endian_int),
    ('storageLimit', big_endian_int),
    ('epochHeight', big_endian_int),
    ('chainId', big_endian_int),
    ('data', binary),
)

class UnsignedTransaction(HashableRLP):
    fields = UNSIGNED_TRANSACTION_FIELDS

class Transaction(HashableRLP):
    fields = (
        ('tx_meta', UnsignedTransaction),
        ('v', big_endian_int),
        ('r', big_endian_int),
        ('s', big_endian_int),
    )


def serializable_unsigned_transaction_from_dict(transaction_dict: TxDict) -> UnsignedTransaction:
    assert_valid_fields(transaction_dict)
    filled_transaction = pipe(
        transaction_dict,
        dict,
        partial(merge, TRANSACTION_DEFAULTS),
        hexstr_if_base32,
        apply_formatters_to_dict(TRANSACTION_FORMATTERS),
    )
    serializer = UnsignedTransaction
    return serializer.from_dict(filled_transaction)

def encode_transaction(unsigned_transaction: UnsignedTransaction, vrs: Tuple[int, int, int]) -> bytes:
    (v, r, s) = vrs
    chain_naive_transaction = dissoc(unsigned_transaction.as_dict(), 'v', 'r', 's')
    signed_transaction = Transaction(tx_meta=UnsignedTransaction(**chain_naive_transaction), v=v, r=r, s=s)
    return rlp.encode(signed_transaction)

def hexstr_if_base32(transaction_dict: TxDict) -> TxDict:
    to = transaction_dict.get("to", None)
    if not (to in VALID_EMPTY_ADDRESSES):
        address = Base32Address(to) # type: ignore
        transaction_dict['to'] = address.hex_address
    return transaction_dict

def is_empty_or_valid_base32_address(val: Any) -> bool:
    if val in VALID_EMPTY_ADDRESSES:
        return True
    else:
        return Base32Address.is_valid_base32(val)

TRANSACTION_FORMATTERS = {
    'nonce': hexstr_if_str(to_int),
    'gasPrice': hexstr_if_str(to_int),
    'gas': hexstr_if_str(to_int),
    'to': apply_one_of_formatters((
        (is_string, hexstr_if_str(to_bytes)),
        (is_bytes, identity),
        (is_none, lambda val: b''),
    )),
    'value': hexstr_if_str(to_int),
    'storageLimit': hexstr_if_str(to_int),
    'epochHeight': hexstr_if_str(to_int),
    'chainId': hexstr_if_str(to_int),
    'data': hexstr_if_str(to_bytes),
    'v': hexstr_if_str(to_int),
    'r': hexstr_if_str(to_int),
    's': hexstr_if_str(to_int),
}

TRANSACTION_VALID_VALUES = {
    'nonce': is_int_or_prefixed_hexstr,
    'gasPrice': is_int_or_prefixed_hexstr,
    'gas': is_int_or_prefixed_hexstr,
    'to': is_empty_or_valid_base32_address,
    'value': is_int_or_prefixed_hexstr,
    'storageLimit': is_int_or_prefixed_hexstr,
    'epochHeight': is_int_or_prefixed_hexstr,
    'chainId': is_int_or_prefixed_hexstr,
    'data': lambda val: isinstance(val, (int, str, bytes, bytearray)),
}

ALLOWED_TRANSACTION_KEYS = {
    'nonce',
    'gasPrice',
    'gas',
    'to',
    'value',
    'storageLimit',
    'epochHeight',
    'chainId',
    'data',
}

REQUIRED_TRANSACITON_KEYS = ALLOWED_TRANSACTION_KEYS.difference(TRANSACTION_DEFAULTS.keys())

def assert_valid_fields(transaction_dict: Any):
    # check if any keys are missing
    missing_keys = REQUIRED_TRANSACITON_KEYS.difference(transaction_dict.keys())
    if missing_keys:
        raise TypeError("Transaction must include these fields: %r" % missing_keys)

    # check if any extra keys were specified
    superfluous_keys = set(transaction_dict.keys()).difference(ALLOWED_TRANSACTION_KEYS)
    if superfluous_keys:
        raise TypeError("Transaction must not include unrecognized fields: %r" % superfluous_keys)

    # check for valid types in each field
    valid_fields: Dict[str, bool]
    valid_fields = apply_formatters_to_dict(TRANSACTION_VALID_VALUES, transaction_dict)
    if not all(valid_fields.values()):
        invalid = {key: transaction_dict[key] for key, valid in valid_fields.items() if not valid}
        raise TypeError("Transaction had invalid fields: %r" % invalid)

def vrs_from(transaction: Transaction) -> Tuple[int, int, int]:
    return (transaction.v, transaction.r, transaction.s) # type: ignore



from typing import TYPE_CHECKING, Optional, Union, Type
from eth_account.signers.local import LocalAccount as EthLocalAccount
from eth_account.datastructures import (
    # SignedMessage,
    SignedTransaction,
)
from cfx_address import (
    eth_eoa_address_to_cfx_hex,
    Base32Address
)
from cfx_address.utils import (
    validate_network_id
)
from cfx_utils.types import (
    TxParam,
    ChecksumAddress,
    HexAddress,
)

if TYPE_CHECKING:
    from cfx_account import Account


class LocalAccount(EthLocalAccount):
    def __init__(self, key, account: Union["Account", Type["Account"]], network_id: Optional[int]=None):
        if network_id is not None:
            validate_network_id(network_id)
        self._network_id = network_id
        
        super().__init__(key, account)

    @property
    def network_id(self) -> Union[int, None]:
        return self._network_id
    
    @network_id.setter
    def network_id(self, new_network_id: Union[int, None]):
        if new_network_id is not None:
            validate_network_id(new_network_id)
        self._network_id = new_network_id
        
    # def reset_network_id(self):
    #     self._network_id = None

    @property
    def address(self) -> Union[Base32Address, HexAddress]:
        """
        returns the address of the account

        :return Union[Base32Address, HexAddress]: _description_
        """
        hex_address = self.hex_address
        if not self._network_id:
            return hex_address
        return Base32Address(hex_address, self._network_id)

    @property
    def hex_address(self) -> HexAddress:
        return eth_eoa_address_to_cfx_hex(super().address)
    
    def get_base32_address(self, specific_network_id: int) -> Base32Address:
        return Base32Address(self.address, specific_network_id)
    
    # add type hint
    def sign_transaction(self, transaction_dict: TxParam) -> SignedTransaction:
        return super().sign_transaction(transaction_dict)
